/********************************************************************

Kinebick - experimental software symmetric block cipher

Runs with 256..4096 bit key and 128 bit block.

Edition 1.0 dated: 15.03.2009

(C) 2009 Alexander Myasnikov, public domain

Web: www.alexanderwdark.narod.ru

********************************************************************/

#include <stdlib.h>
#include <string.h>

#ifdef PRNG_TEST
#include <stdio.h>
#endif

typedef unsigned char u8;
typedef unsigned long u32;
typedef unsigned short u16;

/************** Full key size 256 - 4096 ***************************/
#define keysizebits  512
/*******************************************************************/

/************** Enable lite and fast mode **************************/
#undef LITE_MODE
/*******************************************************************/

static const u32 keysizebytes = keysizebits / 8;
static const u32 keysize = keysizebytes / 4;

/************** Sbox type 1 - 7 ************************************/
#define sizeshift 5
/*******************************************************************/

/************** How to shift in F function  ************************/
#define SHLVAL 6
#define SHRVAL 8
/*******************************************************************/

/************** Optimal constants for key expand  ******************/
#define delta  0xb7e15163
#define cdelta 0x9e3779b9
/*******************************************************************/

/************** How to expand key **********************************/
#define USE_EXTRA_KS
/*******************************************************************/

#ifdef LITE_MODE

#define rounds 64

#else

#define rounds 128

#endif

static const u32 sboxsize = (2 << sizeshift);

#include "sbox.h"

#define ROL(x,y) _lrotl((x), (y))
#define ROR(x,y) _lrotr((x), (y))

#define subst(x,a,b,c,d, r) (ROL((sbox[a][(x) >> 24 & 255] << 24 | sbox[b][(x) >> 16 & 255] << 16 |  sbox[c][(x) >> 8 & 255] << 8 | sbox[d][(x) & 255]),(r)))

#define BSWAP(x) (((x) << 16) | ((x) >> 16))

#define FQ(x, y) ((x ^ y) + ((x << SHLVAL) ^ (x >> SHRVAL)))
#define F(x) (BSWAP(x) + ((x << SHLVAL) ^ (x >> SHRVAL)))
#define byte(x,n)   ((u8)((x) >> (8 * n)))

u32 roundkeys[rounds], _roundkeys[rounds];

#include "native.h"

u32 rawkey[8];


void __stdcall
setup (u32 * key)
{

#ifdef USE_EXTRA_KS
	u8 *subkeys = (u8 *) & roundkeys[0];
	u8 *material = (u8 *) & _roundkeys[0];
	u8 tmp;
#endif

	u32 cnt = delta, kcnt = cdelta, i, n;
	u32 rk[keysize];


	for (i = 0; i != keysize; i++)
	{
		rk[i] = key[i];
	}

	for (i = 0; i != rounds; i++)
	{

		_roundkeys[i] = rk[i % keysize];
		kcnt += rk[i % keysize] + cdelta;
		_roundkeys[i] += cnt;
		_roundkeys[i] =
			subst (_roundkeys[i], i % sboxsize, cnt % sboxsize,
			_roundkeys[i] % sboxsize, kcnt % sboxsize, 11);
		kcnt += _roundkeys[i];
		_roundkeys[i] = F (_roundkeys[i]) + rk[i % keysize];
		_roundkeys[i] =
			subst (_roundkeys[i], kcnt % sboxsize, _roundkeys[i] % sboxsize,
			cnt % sboxsize, i % sboxsize, 11);
		kcnt += _roundkeys[i];
		cnt += delta;

	}

#ifdef USE_EXTRA_KS

	cnt = 0;

	for (n = 0; n != rounds >> 6; n++)
	{

		for (i = 0; i != 256; i++)
		{
			subkeys[i] = i;
		}

		for (i = 0; i != 768; i++)
		{
			cnt = subkeys[(cnt + subkeys[i & 0xff] + material[i % 256]) & 0xff];
			tmp = subkeys[i & 0xff];
			subkeys[i & 0xff] = subkeys[cnt & 0xff];
			subkeys[cnt & 0xff] = tmp;
		}

		subkeys += 0x100;
		material += 0x100;
	}

#else

	for (i = 0; i != rounds; i++)
	{

		roundkeys[i] = _roundkeys[i];
	}

#endif



	for (i = 0; i != 8; i++)
	{
		rawkey[i] = rk[i];
	}

	if (keysize >= 16)

	{
		for (i = 8; i != 16; i++)

		{

			rawkey[i % 8] += rk[i];

		}

	}

	if (keysize > 16)

	{
		for (i = 16; i != keysize; i++)

		{

			rawkey[i % 8] ^= rk[i];

		}
	}


}

#define f0(x,y,z)  (x^y^z)
#define f48(x,y,z) ((y)^((z)&((x)^(y))))
#define f16(x,y,z) ((z)^((x)&((y)^(z))))
#define f32(x,y,z) ((x|~(y))^z)

#define subRound128(a, b, c, d, f, k, r, keyval)  ( a = ROL(a + f(b,c,d) + keyval + k, r)  )

#define backRound128(a, b, c, d, f, k, r, keyval)  ( a = ROR(a, r) - f(b,c,d) - keyval - k  )


void __stdcall
crypt (u32 * data)
{
	u32 register A, B, C, D;

	A = data[0] + rawkey[0];
	B = data[1] + rawkey[1];
	C = data[2] + rawkey[2];
	D = data[3] + rawkey[3];

	subRound128 (A, B, C, D, f0, K[0], 11, roundkeys[0]);
	subRound128 (D, A, B, C, f0, K[1], 14, roundkeys[1]);
	subRound128 (C, D, A, B, f0, K[2], 15, roundkeys[2]);
	subRound128 (B, C, D, A, f0, K[3], 12, roundkeys[3]);
	subRound128 (A, B, C, D, f0, K[4], 5, roundkeys[4]);
	subRound128 (D, A, B, C, f0, K[5], 8, roundkeys[5]);
	subRound128 (C, D, A, B, f0, K[6], 7, roundkeys[6]);
	subRound128 (B, C, D, A, f0, K[7], 9, roundkeys[7]);
	subRound128 (A, B, C, D, f0, K[8], 11, roundkeys[8]);
	subRound128 (D, A, B, C, f0, K[9], 13, roundkeys[9]);
	subRound128 (C, D, A, B, f0, K[10], 14, roundkeys[10]);
	subRound128 (B, C, D, A, f0, K[11], 15, roundkeys[11]);
	subRound128 (A, B, C, D, f0, K[12], 6, roundkeys[12]);
	subRound128 (D, A, B, C, f0, K[13], 7, roundkeys[13]);
	subRound128 (C, D, A, B, f0, K[14], 9, roundkeys[14]);
	subRound128 (B, C, D, A, f0, K[15], 8, roundkeys[15]);

#ifndef LITE_MODE

	subRound128 (A, B, C, D, f48, KK[0], 8, roundkeys[64]);
	subRound128 (D, A, B, C, f48, KK[1], 9, roundkeys[65]);
	subRound128 (C, D, A, B, f48, KK[2], 9, roundkeys[66]);
	subRound128 (B, C, D, A, f48, KK[3], 11, roundkeys[67]);
	subRound128 (A, B, C, D, f48, KK[4], 13, roundkeys[68]);
	subRound128 (D, A, B, C, f48, KK[5], 15, roundkeys[69]);
	subRound128 (C, D, A, B, f48, KK[6], 15, roundkeys[70]);
	subRound128 (B, C, D, A, f48, KK[7], 5, roundkeys[71]);
	subRound128 (A, B, C, D, f48, KK[8], 7, roundkeys[72]);
	subRound128 (D, A, B, C, f48, KK[9], 7, roundkeys[73]);
	subRound128 (C, D, A, B, f48, KK[10], 8, roundkeys[74]);
	subRound128 (B, C, D, A, f48, KK[11], 11, roundkeys[75]);
	subRound128 (A, B, C, D, f48, KK[12], 14, roundkeys[76]);
	subRound128 (D, A, B, C, f48, KK[13], 14, roundkeys[77]);
	subRound128 (C, D, A, B, f48, KK[14], 12, roundkeys[78]);
	subRound128 (B, C, D, A, f48, KK[15], 6, roundkeys[79]);

#endif


	subRound128 (A, B, C, D, f16, K[16], 7, roundkeys[16]);
	subRound128 (D, A, B, C, f16, K[17], 6, roundkeys[17]);
	subRound128 (C, D, A, B, f16, K[18], 8, roundkeys[18]);
	subRound128 (B, C, D, A, f16, K[19], 13, roundkeys[19]);
	subRound128 (A, B, C, D, f16, K[20], 11, roundkeys[20]);
	subRound128 (D, A, B, C, f16, K[21], 9, roundkeys[21]);
	subRound128 (C, D, A, B, f16, K[22], 7, roundkeys[22]);
	subRound128 (B, C, D, A, f16, K[23], 15, roundkeys[23]);
	subRound128 (A, B, C, D, f16, K[24], 7, roundkeys[24]);
	subRound128 (D, A, B, C, f16, K[25], 12, roundkeys[25]);
	subRound128 (C, D, A, B, f16, K[26], 15, roundkeys[26]);
	subRound128 (B, C, D, A, f16, K[27], 9, roundkeys[27]);
	subRound128 (A, B, C, D, f16, K[28], 11, roundkeys[28]);
	subRound128 (D, A, B, C, f16, K[29], 7, roundkeys[29]);
	subRound128 (C, D, A, B, f16, K[30], 13, roundkeys[30]);
	subRound128 (B, C, D, A, f16, K[31], 12, roundkeys[31]);


#ifndef LITE_MODE

	subRound128 (A, B, C, D, f32, KK[16], 9, roundkeys[80]);
	subRound128 (D, A, B, C, f32, KK[17], 13, roundkeys[81]);
	subRound128 (C, D, A, B, f32, KK[18], 15, roundkeys[82]);
	subRound128 (B, C, D, A, f32, KK[19], 7, roundkeys[83]);
	subRound128 (A, B, C, D, f32, KK[20], 12, roundkeys[84]);
	subRound128 (D, A, B, C, f32, KK[21], 8, roundkeys[85]);
	subRound128 (C, D, A, B, f32, KK[22], 9, roundkeys[86]);
	subRound128 (B, C, D, A, f32, KK[23], 11, roundkeys[87]);
	subRound128 (A, B, C, D, f32, KK[24], 7, roundkeys[88]);
	subRound128 (D, A, B, C, f32, KK[25], 7, roundkeys[89]);
	subRound128 (C, D, A, B, f32, KK[26], 12, roundkeys[90]);
	subRound128 (B, C, D, A, f32, KK[27], 7, roundkeys[91]);
	subRound128 (A, B, C, D, f32, KK[28], 6, roundkeys[92]);
	subRound128 (D, A, B, C, f32, KK[29], 15, roundkeys[93]);
	subRound128 (C, D, A, B, f32, KK[30], 13, roundkeys[94]);
	subRound128 (B, C, D, A, f32, KK[31], 11, roundkeys[95]);

#endif

	subRound128 (A, B, C, D, f32, K[32], 11, roundkeys[32]);
	subRound128 (D, A, B, C, f32, K[33], 13, roundkeys[33]);
	subRound128 (C, D, A, B, f32, K[34], 6, roundkeys[34]);
	subRound128 (B, C, D, A, f32, K[35], 7, roundkeys[35]);
	subRound128 (A, B, C, D, f32, K[36], 14, roundkeys[36]);
	subRound128 (D, A, B, C, f32, K[37], 9, roundkeys[37]);
	subRound128 (C, D, A, B, f32, K[38], 13, roundkeys[38]);
	subRound128 (B, C, D, A, f32, K[39], 15, roundkeys[39]);
	subRound128 (A, B, C, D, f32, K[40], 14, roundkeys[40]);
	subRound128 (D, A, B, C, f32, K[41], 8, roundkeys[41]);
	subRound128 (C, D, A, B, f32, K[42], 13, roundkeys[42]);
	subRound128 (B, C, D, A, f32, K[43], 6, roundkeys[43]);
	subRound128 (A, B, C, D, f32, K[44], 5, roundkeys[44]);
	subRound128 (D, A, B, C, f32, K[45], 12, roundkeys[45]);
	subRound128 (C, D, A, B, f32, K[46], 7, roundkeys[46]);
	subRound128 (B, C, D, A, f32, K[47], 5, roundkeys[47]);

#ifndef LITE_MODE

	subRound128 (A, B, C, D, f16, KK[32], 9, roundkeys[96]);
	subRound128 (D, A, B, C, f16, KK[33], 7, roundkeys[97]);
	subRound128 (C, D, A, B, f16, KK[34], 15, roundkeys[98]);
	subRound128 (B, C, D, A, f16, KK[35], 11, roundkeys[99]);
	subRound128 (A, B, C, D, f16, KK[36], 8, roundkeys[100]);
	subRound128 (D, A, B, C, f16, KK[37], 6, roundkeys[101]);
	subRound128 (C, D, A, B, f16, KK[38], 6, roundkeys[102]);
	subRound128 (B, C, D, A, f16, KK[39], 14, roundkeys[103]);
	subRound128 (A, B, C, D, f16, KK[40], 12, roundkeys[104]);
	subRound128 (D, A, B, C, f16, KK[41], 13, roundkeys[105]);
	subRound128 (C, D, A, B, f16, KK[42], 5, roundkeys[106]);
	subRound128 (B, C, D, A, f16, KK[43], 14, roundkeys[107]);
	subRound128 (A, B, C, D, f16, KK[44], 13, roundkeys[108]);
	subRound128 (D, A, B, C, f16, KK[45], 13, roundkeys[109]);
	subRound128 (C, D, A, B, f16, KK[46], 7, roundkeys[110]);
	subRound128 (B, C, D, A, f16, KK[47], 5, roundkeys[111]);

#endif

	subRound128 (A, B, C, D, f48, K[48], 11, roundkeys[48]);
	subRound128 (D, A, B, C, f48, K[49], 12, roundkeys[49]);
	subRound128 (C, D, A, B, f48, K[50], 14, roundkeys[50]);
	subRound128 (B, C, D, A, f48, K[51], 15, roundkeys[51]);
	subRound128 (A, B, C, D, f48, K[52], 14, roundkeys[52]);
	subRound128 (D, A, B, C, f48, K[53], 15, roundkeys[53]);
	subRound128 (C, D, A, B, f48, K[54], 9, roundkeys[54]);
	subRound128 (B, C, D, A, f48, K[55], 8, roundkeys[55]);
	subRound128 (A, B, C, D, f48, K[56], 9, roundkeys[56]);
	subRound128 (D, A, B, C, f48, K[57], 14, roundkeys[57]);
	subRound128 (C, D, A, B, f48, K[58], 5, roundkeys[58]);
	subRound128 (B, C, D, A, f48, K[59], 6, roundkeys[59]);
	subRound128 (A, B, C, D, f48, K[60], 8, roundkeys[60]);
	subRound128 (D, A, B, C, f48, K[62], 6, roundkeys[61]);
	subRound128 (C, D, A, B, f48, K[62], 5, roundkeys[62]);
	subRound128 (B, C, D, A, f48, K[63], 12, roundkeys[63]);

#ifndef LITE_MODE

	subRound128 (A, B, C, D, f0, KK[48], 15, roundkeys[112]);
	subRound128 (D, A, B, C, f0, KK[49], 5, roundkeys[113]);
	subRound128 (C, D, A, B, f0, KK[50], 8, roundkeys[114]);
	subRound128 (B, C, D, A, f0, KK[51], 11, roundkeys[115]);
	subRound128 (A, B, C, D, f0, KK[52], 14, roundkeys[116]);
	subRound128 (D, A, B, C, f0, KK[53], 14, roundkeys[117]);
	subRound128 (C, D, A, B, f0, KK[54], 6, roundkeys[118]);
	subRound128 (B, C, D, A, f0, KK[55], 14, roundkeys[119]);
	subRound128 (A, B, C, D, f0, KK[56], 6, roundkeys[120]);
	subRound128 (D, A, B, C, f0, KK[57], 9, roundkeys[121]);
	subRound128 (C, D, A, B, f0, KK[58], 12, roundkeys[112]);
	subRound128 (B, C, D, A, f0, KK[59], 9, roundkeys[123]);
	subRound128 (A, B, C, D, f0, KK[60], 12, roundkeys[124]);
	subRound128 (D, A, B, C, f0, KK[61], 5, roundkeys[125]);
	subRound128 (C, D, A, B, f0, KK[62], 15, roundkeys[126]);
	subRound128 (B, C, D, A, f0, KK[63], 8, roundkeys[127]);

#endif

	data[0] = A ^ rawkey[4];
	data[1] = B ^ rawkey[5];
	data[2] = C ^ rawkey[6];
	data[3] = D ^ rawkey[7];


}


void __stdcall
decrypt (u32 * data)
{
	u32 register A, B, C, D;

	A = data[0] ^ rawkey[4];
	B = data[1] ^ rawkey[5];
	C = data[2] ^ rawkey[6];
	D = data[3] ^ rawkey[7];

#ifndef LITE_MODE

	backRound128 (B, C, D, A, f0, KK[63], 8, roundkeys[127]);
	backRound128 (C, D, A, B, f0, KK[62], 15, roundkeys[126]);
	backRound128 (D, A, B, C, f0, KK[61], 5, roundkeys[125]);
	backRound128 (A, B, C, D, f0, KK[60], 12, roundkeys[124]);
	backRound128 (B, C, D, A, f0, KK[59], 9, roundkeys[123]);
	backRound128 (C, D, A, B, f0, KK[58], 12, roundkeys[112]);
	backRound128 (D, A, B, C, f0, KK[57], 9, roundkeys[121]);
	backRound128 (A, B, C, D, f0, KK[56], 6, roundkeys[120]);
	backRound128 (B, C, D, A, f0, KK[55], 14, roundkeys[119]);
	backRound128 (C, D, A, B, f0, KK[54], 6, roundkeys[118]);
	backRound128 (D, A, B, C, f0, KK[53], 14, roundkeys[117]);
	backRound128 (A, B, C, D, f0, KK[52], 14, roundkeys[116]);
	backRound128 (B, C, D, A, f0, KK[51], 11, roundkeys[115]);
	backRound128 (C, D, A, B, f0, KK[50], 8, roundkeys[114]);
	backRound128 (D, A, B, C, f0, KK[49], 5, roundkeys[113]);
	backRound128 (A, B, C, D, f0, KK[48], 15, roundkeys[112]);

#endif

	backRound128 (B, C, D, A, f48, K[63], 12, roundkeys[63]);
	backRound128 (C, D, A, B, f48, K[62], 5, roundkeys[62]);
	backRound128 (D, A, B, C, f48, K[62], 6, roundkeys[61]);
	backRound128 (A, B, C, D, f48, K[60], 8, roundkeys[60]);
	backRound128 (B, C, D, A, f48, K[59], 6, roundkeys[59]);
	backRound128 (C, D, A, B, f48, K[58], 5, roundkeys[58]);
	backRound128 (D, A, B, C, f48, K[57], 14, roundkeys[57]);
	backRound128 (A, B, C, D, f48, K[56], 9, roundkeys[56]);
	backRound128 (B, C, D, A, f48, K[55], 8, roundkeys[55]);
	backRound128 (C, D, A, B, f48, K[54], 9, roundkeys[54]);
	backRound128 (D, A, B, C, f48, K[53], 15, roundkeys[53]);
	backRound128 (A, B, C, D, f48, K[52], 14, roundkeys[52]);
	backRound128 (B, C, D, A, f48, K[51], 15, roundkeys[51]);
	backRound128 (C, D, A, B, f48, K[50], 14, roundkeys[50]);
	backRound128 (D, A, B, C, f48, K[49], 12, roundkeys[49]);
	backRound128 (A, B, C, D, f48, K[48], 11, roundkeys[48]);

#ifndef LITE_MODE

	backRound128 (B, C, D, A, f16, KK[47], 5, roundkeys[111]);
	backRound128 (C, D, A, B, f16, KK[46], 7, roundkeys[110]);
	backRound128 (D, A, B, C, f16, KK[45], 13, roundkeys[109]);
	backRound128 (A, B, C, D, f16, KK[44], 13, roundkeys[108]);
	backRound128 (B, C, D, A, f16, KK[43], 14, roundkeys[107]);
	backRound128 (C, D, A, B, f16, KK[42], 5, roundkeys[106]);
	backRound128 (D, A, B, C, f16, KK[41], 13, roundkeys[105]);
	backRound128 (A, B, C, D, f16, KK[40], 12, roundkeys[104]);
	backRound128 (B, C, D, A, f16, KK[39], 14, roundkeys[103]);
	backRound128 (C, D, A, B, f16, KK[38], 6, roundkeys[102]);
	backRound128 (D, A, B, C, f16, KK[37], 6, roundkeys[101]);
	backRound128 (A, B, C, D, f16, KK[36], 8, roundkeys[100]);
	backRound128 (B, C, D, A, f16, KK[35], 11, roundkeys[99]);
	backRound128 (C, D, A, B, f16, KK[34], 15, roundkeys[98]);
	backRound128 (D, A, B, C, f16, KK[33], 7, roundkeys[97]);
	backRound128 (A, B, C, D, f16, KK[32], 9, roundkeys[96]);

#endif

	backRound128 (B, C, D, A, f32, K[47], 5, roundkeys[47]);
	backRound128 (C, D, A, B, f32, K[46], 7, roundkeys[46]);
	backRound128 (D, A, B, C, f32, K[45], 12, roundkeys[45]);
	backRound128 (A, B, C, D, f32, K[44], 5, roundkeys[44]);
	backRound128 (B, C, D, A, f32, K[43], 6, roundkeys[43]);
	backRound128 (C, D, A, B, f32, K[42], 13, roundkeys[42]);
	backRound128 (D, A, B, C, f32, K[41], 8, roundkeys[41]);
	backRound128 (A, B, C, D, f32, K[40], 14, roundkeys[40]);
	backRound128 (B, C, D, A, f32, K[39], 15, roundkeys[39]);
	backRound128 (C, D, A, B, f32, K[38], 13, roundkeys[38]);
	backRound128 (D, A, B, C, f32, K[37], 9, roundkeys[37]);
	backRound128 (A, B, C, D, f32, K[36], 14, roundkeys[36]);
	backRound128 (B, C, D, A, f32, K[35], 7, roundkeys[35]);
	backRound128 (C, D, A, B, f32, K[34], 6, roundkeys[34]);
	backRound128 (D, A, B, C, f32, K[33], 13, roundkeys[33]);
	backRound128 (A, B, C, D, f32, K[32], 11, roundkeys[32]);


#ifndef LITE_MODE

	backRound128 (B, C, D, A, f32, KK[31], 11, roundkeys[95]);
	backRound128 (C, D, A, B, f32, KK[30], 13, roundkeys[94]);
	backRound128 (D, A, B, C, f32, KK[29], 15, roundkeys[93]);
	backRound128 (A, B, C, D, f32, KK[28], 6, roundkeys[92]);
	backRound128 (B, C, D, A, f32, KK[27], 7, roundkeys[91]);
	backRound128 (C, D, A, B, f32, KK[26], 12, roundkeys[90]);
	backRound128 (D, A, B, C, f32, KK[25], 7, roundkeys[89]);
	backRound128 (A, B, C, D, f32, KK[24], 7, roundkeys[88]);
	backRound128 (B, C, D, A, f32, KK[23], 11, roundkeys[87]);
	backRound128 (C, D, A, B, f32, KK[22], 9, roundkeys[86]);
	backRound128 (D, A, B, C, f32, KK[21], 8, roundkeys[85]);
	backRound128 (A, B, C, D, f32, KK[20], 12, roundkeys[84]);
	backRound128 (B, C, D, A, f32, KK[19], 7, roundkeys[83]);
	backRound128 (C, D, A, B, f32, KK[18], 15, roundkeys[82]);
	backRound128 (D, A, B, C, f32, KK[17], 13, roundkeys[81]);
	backRound128 (A, B, C, D, f32, KK[16], 9, roundkeys[80]);

#endif


	backRound128 (B, C, D, A, f16, K[31], 12, roundkeys[31]);
	backRound128 (C, D, A, B, f16, K[30], 13, roundkeys[30]);
	backRound128 (D, A, B, C, f16, K[29], 7, roundkeys[29]);
	backRound128 (A, B, C, D, f16, K[28], 11, roundkeys[28]);
	backRound128 (B, C, D, A, f16, K[27], 9, roundkeys[27]);
	backRound128 (C, D, A, B, f16, K[26], 15, roundkeys[26]);
	backRound128 (D, A, B, C, f16, K[25], 12, roundkeys[25]);
	backRound128 (A, B, C, D, f16, K[24], 7, roundkeys[24]);
	backRound128 (B, C, D, A, f16, K[23], 15, roundkeys[23]);
	backRound128 (C, D, A, B, f16, K[22], 7, roundkeys[22]);
	backRound128 (D, A, B, C, f16, K[21], 9, roundkeys[21]);
	backRound128 (A, B, C, D, f16, K[20], 11, roundkeys[20]);
	backRound128 (B, C, D, A, f16, K[19], 13, roundkeys[19]);
	backRound128 (C, D, A, B, f16, K[18], 8, roundkeys[18]);
	backRound128 (D, A, B, C, f16, K[17], 6, roundkeys[17]);
	backRound128 (A, B, C, D, f16, K[16], 7, roundkeys[16]);


#ifndef LITE_MODE

	backRound128 (B, C, D, A, f48, KK[15], 6, roundkeys[79]);
	backRound128 (C, D, A, B, f48, KK[14], 12, roundkeys[78]);
	backRound128 (D, A, B, C, f48, KK[13], 14, roundkeys[77]);
	backRound128 (A, B, C, D, f48, KK[12], 14, roundkeys[76]);
	backRound128 (B, C, D, A, f48, KK[11], 11, roundkeys[75]);
	backRound128 (C, D, A, B, f48, KK[10], 8, roundkeys[74]);
	backRound128 (D, A, B, C, f48, KK[9], 7, roundkeys[73]);
	backRound128 (A, B, C, D, f48, KK[8], 7, roundkeys[72]);
	backRound128 (B, C, D, A, f48, KK[7], 5, roundkeys[71]);
	backRound128 (C, D, A, B, f48, KK[6], 15, roundkeys[70]);
	backRound128 (D, A, B, C, f48, KK[5], 15, roundkeys[69]);
	backRound128 (A, B, C, D, f48, KK[4], 13, roundkeys[68]);
	backRound128 (B, C, D, A, f48, KK[3], 11, roundkeys[67]);
	backRound128 (C, D, A, B, f48, KK[2], 9, roundkeys[66]);
	backRound128 (D, A, B, C, f48, KK[1], 9, roundkeys[65]);
	backRound128 (A, B, C, D, f48, KK[0], 8, roundkeys[64]);

#endif

	backRound128 (B, C, D, A, f0, K[15], 8, roundkeys[15]);
	backRound128 (C, D, A, B, f0, K[14], 9, roundkeys[14]);
	backRound128 (D, A, B, C, f0, K[13], 7, roundkeys[13]);
	backRound128 (A, B, C, D, f0, K[12], 6, roundkeys[12]);
	backRound128 (B, C, D, A, f0, K[11], 15, roundkeys[11]);
	backRound128 (C, D, A, B, f0, K[10], 14, roundkeys[10]);
	backRound128 (D, A, B, C, f0, K[9], 13, roundkeys[9]);
	backRound128 (A, B, C, D, f0, K[8], 11, roundkeys[8]);
	backRound128 (B, C, D, A, f0, K[7], 9, roundkeys[7]);
	backRound128 (C, D, A, B, f0, K[6], 7, roundkeys[6]);
	backRound128 (D, A, B, C, f0, K[5], 8, roundkeys[5]);
	backRound128 (A, B, C, D, f0, K[4], 5, roundkeys[4]);
	backRound128 (B, C, D, A, f0, K[3], 12, roundkeys[3]);
	backRound128 (C, D, A, B, f0, K[2], 15, roundkeys[2]);
	backRound128 (D, A, B, C, f0, K[1], 14, roundkeys[1]);
	backRound128 (A, B, C, D, f0, K[0], 11, roundkeys[0]);

	data[0] = A - rawkey[0];
	data[1] = B - rawkey[1];
	data[2] = C - rawkey[2];
	data[3] = D - rawkey[3];

}


#ifdef PRNG_TEST

u32 buf[4] = { 0xC4A60B29, 0x1DF6907A, 0xE5F8D137, 0x86C02E59 };
u32 cbuf[4] = { 0xBFE60D0F, 0x7CCB60D6, 0x926A9747, 0x17EE621F };
u32 key[16] =
{ 0x58819022, 0xAC3F8B9D, 0x9CEDAECB, 0x89EA9D3D, 0xB43CD725, 0xF04F4E02,
0x12DF1FF7, 0xDBED2295, 0x9140703B, 0x63C99153, 0x786F5204, 0xF0DEA408,
0x3BE2517F, 0xEB12063E, 0xBEC1F2A8, 0x3A498DC2
};

int
main (int argc, char *argv[])
{
	FILE *f;
	int i, n;

	if (argc < 2)
	{
		fprintf (stderr, "usage: %s <output filename>\n", argv[0]);
		return (1);
	}
	if ((f = fopen (argv[1], "wb+")) == NULL)
	{
		fprintf (stderr, "failed to open '%s' for writing.\n", argv[1]);
		return (1);
	}



	setup (&key[0]);
	for (i = 0; i != 67108864; i++)
	{

		crypt (&buf[0]);

		for (n = 0; n != 4; n++)
		{
			cbuf[n] ^= buf[n];
		}

		fwrite (cbuf, 16, 1, f);


	}

	fclose (f);
	return (0);
}

#endif



u32 __stdcall getblocksize ()
{
	return 128;
}

u32 __stdcall getkeysize ()
{
	return keysizebits;
}

void __stdcall getciphername (unsigned char *p)
{
strcpy(p, "Kinebick");
}