/********************************************************************

Shogashi - experimental software symmetric block cipher

Runs with 256..4096 bit key and 128 bit block.

Edition 1.0 dated: 20.03.2009

(C) 2009 Alexander Myasnikov, public domain

Web: www.alexanderwdark.narod.ru

********************************************************************/

#include <stdlib.h>
#include <string.h>

#ifdef PRNG_TEST
#include <stdio.h>
#endif

typedef unsigned char u8;
typedef unsigned long u32;
typedef unsigned short u16;

/************** Full key size 256 - 4096 ***************************/
#define keysizebits  512
/*******************************************************************/

static const u32 keysizebytes = keysizebits / 8;
static const u32 keysize = keysizebytes / 4;

/************** Sbox type 1 - 7 ************************************/
#define sizeshift 6
/*******************************************************************/

/************** How to shift in F function  ************************/
#define SHLVAL 6
#define SHRVAL 8
/*******************************************************************/

/************** Optimal constants for key expand  ******************/
#define delta  0xb7e15163
#define cdelta 0x9e3779b9
/*******************************************************************/

/************** How to expand key **********************************/
#define USE_EXTRA_KS
/*******************************************************************/

#define rounds 128

static const u32 sboxsize = (2 << sizeshift);

#include "sbox.h"

#define ROL(x,y) _lrotl((x), (y))
#define ROR(x,y) _lrotr((x), (y))

#define subst(x,a,b,c,d, r) (ROL((sbox[a][(x) >> 24 & 255] << 24 | sbox[b][(x) >> 16 & 255] << 16 |  sbox[c][(x) >> 8 & 255] << 8 | sbox[d][(x) & 255]),(r)))

#define BSWAP(x) (((x) << 16) | ((x) >> 16))

#define FQ(x, y) ((x ^ y) + ((x << SHLVAL) ^ (x >> SHRVAL)))
#define F(x) (BSWAP(x) + ((x << SHLVAL) ^ (x >> SHRVAL)))
#define byte(x,n)   ((u8)((x) >> (8 * n)))

u32 roundkeys[136], _roundkeys[136];

u32 rawkey[8];

u32 l_box[4][256];

u8 s_box[256] = {
  0xe1, 0x42, 0x3e, 0x81, 0x4e, 0x17, 0x9e, 0xfd, 0xb4, 0x3f, 0x2c, 0xda,
  0x31, 0x1e, 0xe0, 0x41, 0xcc, 0xf3, 0x82, 0x7d, 0x7c, 0x12, 0x8e, 0xbb,
  0xe4, 0x58, 0x15, 0xd5, 0x6f, 0xe9, 0x4c, 0x4b, 0x35, 0x7b, 0x5a, 0x9a,
  0x90, 0x45, 0xbc, 0xf8, 0x79, 0xd6, 0x1b, 0x88, 0x02, 0xab, 0xcf, 0x64,
  0x09, 0x0c, 0xf0, 0x01, 0xa4, 0xb0, 0xf6, 0x93, 0x43, 0x63, 0x86, 0xdc,
  0x11, 0xa5, 0x83, 0x8b, 0xc9, 0xd0, 0x19, 0x95, 0x6a, 0xa1, 0x5c, 0x24,
  0x6e, 0x50, 0x21, 0x80, 0x2f, 0xe7, 0x53, 0x0f, 0x91, 0x22, 0x04, 0xed,
  0xa6, 0x48, 0x49, 0x67, 0xec, 0xf7, 0xc0, 0x39, 0xce, 0xf2, 0x2d, 0xbe,
  0x5d, 0x1c, 0xe3, 0x87, 0x07, 0x0d, 0x7a, 0xf4, 0xfb, 0x32, 0xf5, 0x8c,
  0xdb, 0x8f, 0x25, 0x96, 0xa8, 0xea, 0xcd, 0x33, 0x65, 0x54, 0x06, 0x8d,
  0x89, 0x0a, 0x5e, 0xd9, 0x16, 0x0e, 0x71, 0x6c, 0x0b, 0xff, 0x60, 0xd2,
  0x2e, 0xd3, 0xc8, 0x55, 0xc2, 0x23, 0xb7, 0x74, 0xe2, 0x9b, 0xdf, 0x77,
  0x2b, 0xb9, 0x3c, 0x62, 0x13, 0xe5, 0x94, 0x34, 0xb1, 0x27, 0x84, 0x9f,
  0xd7, 0x51, 0x00, 0x61, 0xad, 0x85, 0x73, 0x03, 0x08, 0x40, 0xef, 0x68,
  0xfe, 0x97, 0x1f, 0xde, 0xaf, 0x66, 0xe8, 0xb8, 0xae, 0xbd, 0xb3, 0xeb,
  0xc6, 0x6b, 0x47, 0xa9, 0xd8, 0xa7, 0x72, 0xee, 0x1d, 0x7e, 0xaa, 0xb6,
  0x75, 0xcb, 0xd4, 0x30, 0x69, 0x20, 0x7f, 0x37, 0x5b, 0x9d, 0x78, 0xa3,
  0xf1, 0x76, 0xfa, 0x05, 0x3d, 0x3a, 0x44, 0x57, 0x3b, 0xca, 0xc7, 0x8a,
  0x18, 0x46, 0x9c, 0xbf, 0xba, 0x38, 0x56, 0x1a, 0x92, 0x4d, 0x26, 0x29,
  0xa2, 0x98, 0x10, 0x99, 0x70, 0xa0, 0xc5, 0x28, 0xc1, 0x6d, 0x14, 0xac,
  0xf9, 0x5f, 0x4f, 0xc4, 0xc3, 0xd1, 0xfc, 0xdd, 0xb2, 0x59, 0xe6, 0xb5,
  0x36, 0x52, 0x4a, 0x2a
};

u32
mod_inv (u32 x)
{
  u32 y1, y2, a, b, q;

  y1 = ~((-x) / x);
  y2 = 1;

  a = x;
  b = y1 * x;

  for (;;)
  {
    q = a / b;

    if ((a -= q * b) == 0)

      return (x * y1 == 1 ? y1 : -y1);

    y2 -= q * y1;

    q = b / a;

    if ((b -= q * a) == 0)

      return (x * y2 == 1 ? y2 : -y2);

    y1 -= q * y2;
  }
};


void __stdcall
setup (u32 * key)
{

#ifdef USE_EXTRA_KS
  u8 *subkeys = (u8 *) & roundkeys[0];
  u8 *material = (u8 *) & _roundkeys[0];
  u8 tmp;
#endif

  u32 cnt = delta, kcnt = cdelta, i, n;
  u32 rk[keysize];


  for (i = 0; i != keysize; i++)
  {
    rk[i] = key[i];
  }

  for (i = 0; i != rounds; i++)
  {

    _roundkeys[i] = rk[i % keysize];
    kcnt += rk[i % keysize] + cdelta;
    _roundkeys[i] += cnt;
    _roundkeys[i] =
      subst (_roundkeys[i], i % sboxsize, cnt % sboxsize,
      _roundkeys[i] % sboxsize, kcnt % sboxsize, 11);
    kcnt += _roundkeys[i];
    _roundkeys[i] = F (_roundkeys[i]) + rk[i % keysize];
    _roundkeys[i] =
      subst (_roundkeys[i], kcnt % sboxsize, _roundkeys[i] % sboxsize,
      cnt % sboxsize, i % sboxsize, 11);
    kcnt += _roundkeys[i];
    cnt += delta;

  }

#ifdef USE_EXTRA_KS

  cnt = 0;

  for (n = 0; n != rounds >> 6; n++)
  {

    for (i = 0; i != 256; i++)
    {
      subkeys[i] = i;
    }

    for (i = 0; i != 768; i++)
    {
      cnt = subkeys[(cnt + subkeys[i & 0xff] + material[i % 256]) & 0xff];
      tmp = subkeys[i & 0xff];
      subkeys[i & 0xff] = subkeys[cnt & 0xff];
      subkeys[cnt & 0xff] = tmp;
    }

    subkeys += 0x100;
    material += 0x100;
  }

#else

  for (i = 0; i != rounds; i++)
  {

    roundkeys[i] = _roundkeys[i];
  }

#endif



  for (i = 0; i != 8; i++)
  {
    rawkey[i] = rk[i];
  }

  if (keysize >= 16)

  {
    for (i = 8; i != 16; i++)

    {
      rawkey[i % 8] += rk[i];
    }

  }

  if (keysize > 16)

  {
    for (i = 16; i != keysize; i++)

    {
      rawkey[i % 8] ^= rk[i];
    }
  }


  for (i = 0; i < 256; ++i)
  {
    l_box[0][i] = ((u32) (s_box[i]));
    l_box[1][i] = ((u32) (s_box[i])) << 8;
    l_box[2][i] = ((u32) (s_box[i])) << 16;
    l_box[3][i] = ((u32) (s_box[i])) << 24;
  }

  for (i = 104; i < 112; ++i)
  {
    roundkeys[i] |= 1;
    roundkeys[i + 12] = mod_inv (roundkeys[i]);
  }
}

#define bp_fun(a,b,c,d,e,f,g,h)     \
  u = (e ^ g) & 0x00ffff00;         \
  v = (f ^ h) & 0x0000ffff;         \
  a = e ^ u; c = g ^ u;             \
  b = f ^ v; d = h ^ v

#define ibp_fun(a,b,c,d,e,f,g,h)    \
  u = (e ^ g) & 0xff0000ff;         \
  v = (f ^ h) & 0xffff0000;         \
  a = e ^ u; c = g ^ u;             \
  b = f ^ v; d = h ^ v

#define s_fun(x,y)        \
  p = x; q = x >> 8;      \
  r = y; s = y >> 8;      \
  x  = l_box[0][r & 255]; \
  y  = l_box[0][p & 255]; \
  p >>=  16; r >>=  16;   \
  x |= l_box[1][q & 255]; \
  y |= l_box[1][s & 255]; \
  x |= l_box[2][r & 255]; \
  y |= l_box[2][p & 255]; \
  x |= l_box[3][p >> 8];  \
  y |= l_box[3][r >> 8]

#define sx_fun(x,y)       \
  p = x >>  8;            \
  q = x >> 16;            \
  x  = l_box[0][x & 255]; \
  x |= l_box[1][p & 255]; \
  x |= l_box[2][q & 255]; \
  x |= l_box[3][q >> 8];  \
  p = y >>  8;            \
  q = y >> 16;            \
  y  = l_box[0][y & 255]; \
  y |= l_box[1][p & 255]; \
  y |= l_box[2][q & 255]; \
  y |= l_box[3][q >> 8]

#define sp_fun(x,y)       \
  s_fun(x,y);             \
  y ^= x;                 \
  x ^= ROR(y, 16);        \
  y ^= ROR(x, 8);         \
  x ^= y

#define sr_fun(x,y)       \
  p = x; q = x >> 8;      \
  r = y; s = y >> 8;      \
  y  = l_box[1][p & 255]; \
  x  = l_box[1][r & 255]; \
  p >>= 16; r >>= 16;     \
  x |= l_box[2][q & 255]; \
  y |= l_box[2][s & 255]; \
  y |= l_box[3][p & 255]; \
  x |= l_box[3][r & 255]; \
  x |= l_box[0][r >>  8]; \
  y |= l_box[0][p >>  8]

#define f_fun(a,b,c,d,k,i)            \
  u = c ^ (k[i]); v = d ^ (k[i + 1]); \
  sp_fun(u, v);                       \
  u ^= (k[i+ 2]); v ^= (k[i + 3]);    \
  sr_fun(u, v);                       \
  a ^= v;                             \
  b ^= u


void __stdcall
crypt (u32 * data)
{
  register u32 a, b, c, d;
  u32 p, q, r, s, u, v;

  p = data[0];
  q = data[1];
  r = data[2];
  s = data[3];

  p ^= roundkeys[96];
  q ^= roundkeys[97];
  r ^= roundkeys[98];
  s ^= roundkeys[99];
  p *= roundkeys[104];
  q *= roundkeys[105];
  r *= roundkeys[106];
  s *= roundkeys[107];

  p += F (q) + roundkeys[124];
  q += F (r) + roundkeys[125];
  r += F (s) + roundkeys[126];
  s += F (p) + roundkeys[127];

  bp_fun (a, b, c, d, p, q, r, s);

  f_fun (a, b, c, d, roundkeys, 0);
  f_fun (c, d, a, b, roundkeys, 4);
  f_fun (a, b, c, d, roundkeys, 8);
  f_fun (c, d, a, b, roundkeys, 12);
  f_fun (a, b, c, d, roundkeys, 16);
  f_fun (c, d, a, b, roundkeys, 20);
  f_fun (a, b, c, d, roundkeys, 24);
  f_fun (c, d, a, b, roundkeys, 28);
  f_fun (a, b, c, d, roundkeys, 32);
  f_fun (c, d, a, b, roundkeys, 36);
  f_fun (a, b, c, d, roundkeys, 40);
  f_fun (c, d, a, b, roundkeys, 44);
  f_fun (a, b, c, d, roundkeys, 48);
  f_fun (c, d, a, b, roundkeys, 52);
  f_fun (a, b, c, d, roundkeys, 56);
  f_fun (c, d, a, b, roundkeys, 60);
  f_fun (a, b, c, d, roundkeys, 64);
  f_fun (c, d, a, b, roundkeys, 68);
  f_fun (a, b, c, d, roundkeys, 72);
  f_fun (c, d, a, b, roundkeys, 76);
  f_fun (a, b, c, d, roundkeys, 80);
  f_fun (c, d, a, b, roundkeys, 84);
  f_fun (a, b, c, d, roundkeys, 88);
  f_fun (c, d, a, b, roundkeys, 92);

  ibp_fun (p, q, r, s, a, b, c, d);

  p *= roundkeys[108];
  q *= roundkeys[109];
  r *= roundkeys[110];
  s *= roundkeys[111];
  p ^= roundkeys[100];
  q ^= roundkeys[101];
  r ^= roundkeys[102];
  s ^= roundkeys[103];

  data[0] = p;
  data[1] = q;
  data[2] = r;
  data[3] = s;
};

void __stdcall
decrypt (u32 * data)
{
  register u32 a, b, c, d;
  u32 p, q, r, s, u, v;

  p = data[0];
  q = data[1];
  r = data[2];
  s = data[3];

  p ^= roundkeys[100];
  q ^= roundkeys[101];
  r ^= roundkeys[102];
  s ^= roundkeys[103];
  p *= roundkeys[120];
  q *= roundkeys[121];
  r *= roundkeys[122];
  s *= roundkeys[123];

  bp_fun (a, b, c, d, p, q, r, s);

  f_fun (a, b, c, d, roundkeys, 92);
  f_fun (c, d, a, b, roundkeys, 88);
  f_fun (a, b, c, d, roundkeys, 84);
  f_fun (c, d, a, b, roundkeys, 80);
  f_fun (a, b, c, d, roundkeys, 76);
  f_fun (c, d, a, b, roundkeys, 72);
  f_fun (a, b, c, d, roundkeys, 68);
  f_fun (c, d, a, b, roundkeys, 64);
  f_fun (a, b, c, d, roundkeys, 60);
  f_fun (c, d, a, b, roundkeys, 56);
  f_fun (a, b, c, d, roundkeys, 52);
  f_fun (c, d, a, b, roundkeys, 48);
  f_fun (a, b, c, d, roundkeys, 44);
  f_fun (c, d, a, b, roundkeys, 40);
  f_fun (a, b, c, d, roundkeys, 36);
  f_fun (c, d, a, b, roundkeys, 32);
  f_fun (a, b, c, d, roundkeys, 28);
  f_fun (c, d, a, b, roundkeys, 24);
  f_fun (a, b, c, d, roundkeys, 20);
  f_fun (c, d, a, b, roundkeys, 16);
  f_fun (a, b, c, d, roundkeys, 12);
  f_fun (c, d, a, b, roundkeys, 8);
  f_fun (a, b, c, d, roundkeys, 4);
  f_fun (c, d, a, b, roundkeys, 0);

  ibp_fun (p, q, r, s, a, b, c, d);

  s -= F (p) + roundkeys[127];
  r -= F (s) + roundkeys[126];
  q -= F (r) + roundkeys[125];
  p -= F (q) + roundkeys[124];

  p *= roundkeys[116];
  q *= roundkeys[117];
  r *= roundkeys[118];
  s *= roundkeys[119];
  p ^= roundkeys[96];
  q ^= roundkeys[97];
  r ^= roundkeys[98];
  s ^= roundkeys[99];

  data[0] = p;
  data[1] = q;
  data[2] = r;
  data[3] = s;
};


#ifdef PRNG_TEST

u32 buf[4] = { 0xC4A60B29, 0x1DF6907A, 0xE5F8D137, 0x86C02E59 };
u32 cbuf[4] = { 0xBFE60D0F, 0x7CCB60D6, 0x926A9747, 0x17EE621F };
u32 key[16] =
{ 0x58819022, 0xAC3F8B9D, 0x9CEDAECB, 0x89EA9D3D, 0xB43CD725, 0xF04F4E02,
0x12DF1FF7, 0xDBED2295, 0x9140703B, 0x63C99153, 0x786F5204, 0xF0DEA408,
0x3BE2517F, 0xEB12063E, 0xBEC1F2A8, 0x3A498DC2
};

int
main (int argc, char *argv[])
{
  FILE *f;
  int i, n;

  if (argc < 2)
  {
    fprintf (stderr, "usage: %s <output filename>\n", argv[0]);
    return (1);
  }
  if ((f = fopen (argv[1], "wb+")) == NULL)
  {
    fprintf (stderr, "failed to open '%s' for writing.\n", argv[1]);
    return (1);
  }



  setup (&key[0]);
  for (i = 0; i != 67108864; i++)
  {

    crypt (&buf[0]);

    for (n = 0; n != 4; n++)
    {
      cbuf[n] ^= buf[n];
    }

    fwrite (cbuf, 16, 1, f);


  }

  fclose (f);
  return (0);
}

#endif



u32 __stdcall getblocksize ()
{
	return 128;
}

u32 __stdcall getkeysize ()
{
	return keysizebits;
}

void __stdcall getciphername (unsigned char *p)
{
strcpy(p, "Shogashi");
}