/********************************************************************

Sonjitege - experimental software symmetric block cipher

Runs with 256..4096 bit key and 128 bit block.

Edition 1.0 dated: 17.03.2009

(C) 2009 Alexander Myasnikov, public domain

Web: www.alexanderwdark.narod.ru

********************************************************************/

#include <stdlib.h>
#include <string.h>

#ifdef PRNG_TEST
#include <stdio.h>
#endif

typedef unsigned char u8;
typedef unsigned long u32;
typedef unsigned short u16;

/************** Full key size 256 - 4096 ***************************/
#define keysizebits  512
/*******************************************************************/

static const u32 keysizebytes = keysizebits / 8;
static const u32 keysize = keysizebytes / 4;

/************** Sbox type 1 - 7 ************************************/
#define sizeshift 5
/*******************************************************************/

/************** How to shift in F function  ************************/
#define SHLVAL 6
#define SHRVAL 8
/*******************************************************************/

/************** Optimal constants for key expand  ******************/
#define delta  0xb7e15163
#define cdelta 0x9e3779b9
/*******************************************************************/

/************** How to expand key **********************************/
#define USE_EXTRA_KS
/*******************************************************************/

#define rounds 128

static const u32 sboxsize = (2 << sizeshift);

#include "sbox.h"

#define ROL(x,y) _lrotl((x), (y))
#define ROR(x,y) _lrotr((x), (y))

#define subst(x,a,b,c,d, r) (ROL((sbox[a][(x) >> 24 & 255] << 24 | sbox[b][(x) >> 16 & 255] << 16 |  sbox[c][(x) >> 8 & 255] << 8 | sbox[d][(x) & 255]),(r)))

#define BSWAP(x) (((x) << 16) | ((x) >> 16))

#define FQ(x, y) ((x ^ y) + ((x << SHLVAL) ^ (x >> SHRVAL)))
#define F(x) (BSWAP(x) + ((x << SHLVAL) ^ (x >> SHRVAL)))
#define byte(x,n)   ((u8)((x) >> (8 * n)))

u32 roundkeys[rounds], _roundkeys[rounds];

u32 rawkey[8];


void __stdcall
setup (u32 * key)
{

#ifdef USE_EXTRA_KS
	u8 *subkeys = (u8 *) & roundkeys[0];
	u8 *material = (u8 *) & _roundkeys[0];
	u8 tmp;
#endif

	u32 cnt = delta, kcnt = cdelta, i, n;
	u32 rk[keysize];


	for (i = 0; i != keysize; i++)
	{
		rk[i] = key[i];
	}

	for (i = 0; i != rounds; i++)
	{

		_roundkeys[i] = rk[i % keysize];
		kcnt += rk[i % keysize] + cdelta;
		_roundkeys[i] += cnt;
		_roundkeys[i] =
			subst (_roundkeys[i], i % sboxsize, cnt % sboxsize,
			_roundkeys[i] % sboxsize, kcnt % sboxsize, 11);
		kcnt += _roundkeys[i];
		_roundkeys[i] = F (_roundkeys[i]) + rk[i % keysize];
		_roundkeys[i] =
			subst (_roundkeys[i], kcnt % sboxsize, _roundkeys[i] % sboxsize,
			cnt % sboxsize, i % sboxsize, 11);
		kcnt += _roundkeys[i];
		cnt += delta;

	}

#ifdef USE_EXTRA_KS

	cnt = 0;

	for (n = 0; n != rounds >> 6; n++)
	{

		for (i = 0; i != 256; i++)
		{
			subkeys[i] = i;
		}

		for (i = 0; i != 768; i++)
		{
			cnt = subkeys[(cnt + subkeys[i & 0xff] + material[i % 256]) & 0xff];
			tmp = subkeys[i & 0xff];
			subkeys[i & 0xff] = subkeys[cnt & 0xff];
			subkeys[cnt & 0xff] = tmp;
		}

		subkeys += 0x100;
		material += 0x100;
	}

#else

	for (i = 0; i != rounds; i++)
	{

		roundkeys[i] = _roundkeys[i];
	}

#endif



	for (i = 0; i != 8; i++)
	{
		rawkey[i] = rk[i];
	}

	if (keysize >= 16)

	{
		for (i = 8; i != 16; i++)

		{

			rawkey[i % 8] += rk[i];

		}

	}

	if (keysize > 16)

	{
		for (i = 16; i != keysize; i++)

		{

			rawkey[i % 8] ^= rk[i];

		}
	}


}

#define f0(b,c,d)   b ^  (c & d)
#define f1(b,c,d)   b ^  (c + d)
#define f2(b,c,d)   b | ~(c & d)
#define f3(b,c,d)   b | ~(c + d)
#define f4(b,c,d)   b &  (c ^ d)
#define f5(b,c,d)   b &  (c + d)
#define f6(b,c,d)   b &  (c | d)
#define f7(b,c,d)   b -  (c & d)
#define f8(b,c,d)   b -  (c + d)
#define f9(b,c,d)   b -  (c ^ d)
#define f10(b,c,d)  b |  ((c + d) & (c ^ d))
#define f11(b,c,d)  b -  ((c + d) | (c ^ d))
#define f12(b,c,d)  b &  ((c + d) | (c ^ d))
#define f13(b,c,d)  b & ~((c & d) ^ (c + d))
#define f14(b,c,d)  b &  ((c | d) + (c ^ d))
#define f15(b,c,d)  b ^  ((c | d) & (c - d))


#define fround(a, b, c, d, func, key, r)  a = ROL(a + (func(b,c,d)) + key, r)

#define bround(a, b, c, d, func, key, r)  a = ROR(a, r) - (func(b,c,d)) - key



void __stdcall
crypt (u32 * data)
{
	u32 register a, b, c, d;

	a = data[0] ^ rawkey[0];
	b = data[1] ^ rawkey[1];
	c = data[2] ^ rawkey[2];
	d = data[3] ^ rawkey[3];

	fround (a, b, c, d, f15, roundkeys[120], 14);
	fround (b, c, a, d, f14, roundkeys[121], 9);
	fround (c, d, a, b, f13, roundkeys[122], 11);
	fround (d, a, b, c, f12, roundkeys[123], 10);
	fround (a, b, d, c, f11, roundkeys[124], 15);
	fround (b, c, d, a, f10, roundkeys[125], 13);
	fround (c, d, b, a, f9, roundkeys[126], 12);
	fround (d, a, c, b, f8, roundkeys[127], 8);

	fround (a, b, c, d, f0, roundkeys[112], 14);
	fround (b, c, a, d, f1, roundkeys[113], 9);
	fround (c, d, a, b, f2, roundkeys[114], 11);
	fround (d, a, b, c, f3, roundkeys[115], 13);
	fround (a, b, d, c, f4, roundkeys[116], 12);
	fround (b, c, d, a, f5, roundkeys[117], 10);
	fround (c, d, b, a, f6, roundkeys[118], 15);
	fround (d, a, c, b, f7, roundkeys[119], 9);

	fround (a, b, c, d, f8, roundkeys[104], 8);
	fround (b, c, a, d, f9, roundkeys[105], 9);
	fround (c, d, a, b, f10, roundkeys[106], 13);
	fround (d, a, b, c, f11, roundkeys[107], 15);
	fround (a, b, d, c, f12, roundkeys[108], 12);
	fround (b, c, d, a, f13, roundkeys[109], 11);
	fround (c, d, b, a, f14, roundkeys[110], 10);
	fround (d, a, c, b, f15, roundkeys[111], 14);

	fround (a, b, c, d, f7, roundkeys[100], 5);
	fround (b, c, a, d, f6, roundkeys[101], 8);
	fround (c, d, a, b, f5, roundkeys[102], 7);
	fround (d, a, b, c, f4, roundkeys[103], 6);

	fround (a, b, c, d, f3, roundkeys[96], 5);
	fround (b, c, a, d, f2, roundkeys[97], 7);
	fround (c, d, a, b, f1, roundkeys[98], 8);
	fround (d, a, b, c, f0, roundkeys[99], 6);

	fround (a, b, c, d, f4, roundkeys[92], 6);
	fround (b, c, a, d, f5, roundkeys[93], 8);
	fround (c, d, a, b, f6, roundkeys[94], 7);
	fround (d, a, b, c, f7, roundkeys[95], 5);

	fround (a, b, c, d, f0, roundkeys[88], 7);
	fround (b, c, a, d, f1, roundkeys[89], 5);
	fround (c, d, a, b, f2, roundkeys[90], 8);
	fround (d, a, b, c, f3, roundkeys[91], 6);

	fround (a, b, c, d, f15, roundkeys[80], 8);
	fround (b, c, a, d, f15, roundkeys[81], 15);
	fround (c, d, a, b, f15, roundkeys[82], 12);
	fround (d, a, b, c, f15, roundkeys[83], 13);
	fround (a, b, d, c, f15, roundkeys[84], 10);
	fround (b, c, d, a, f15, roundkeys[85], 11);
	fround (c, d, b, a, f15, roundkeys[86], 9);
	fround (d, a, c, b, f15, roundkeys[87], 14);

	fround (a, b, c, d, f14, roundkeys[72], 10);
	fround (b, c, a, d, f14, roundkeys[73], 9);
	fround (c, d, a, b, f14, roundkeys[74], 5);
	fround (d, a, b, c, f14, roundkeys[75], 7);
	fround (a, b, d, c, f14, roundkeys[76], 5);
	fround (b, c, d, a, f14, roundkeys[77], 8);
	fround (c, d, b, a, f14, roundkeys[78], 11);
	fround (d, a, c, b, f14, roundkeys[79], 4);


	fround (a, b, c, d, f13, roundkeys[64], 12);
	fround (b, c, a, d, f13, roundkeys[65], 7);
	fround (c, d, a, b, f13, roundkeys[66], 11);
	fround (d, a, b, c, f13, roundkeys[67], 8);
	fround (a, b, d, c, f13, roundkeys[68], 9);
	fround (b, c, d, a, f13, roundkeys[69], 5);
	fround (c, d, b, a, f13, roundkeys[70], 10);
	fround (d, a, c, b, f13, roundkeys[71], 6);

	fround (a, b, c, d, f12, roundkeys[56], 12);
	fround (b, c, a, d, f12, roundkeys[57], 7);
	fround (c, d, a, b, f12, roundkeys[58], 8);
	fround (d, a, b, c, f12, roundkeys[59], 10);
	fround (a, b, d, c, f12, roundkeys[60], 9);
	fround (b, c, d, a, f12, roundkeys[61], 13);
	fround (c, d, b, a, f12, roundkeys[62], 6);
	fround (d, a, c, b, f12, roundkeys[63], 11);

	fround (a, b, c, d, f11, roundkeys[48], 13);
	fround (b, c, a, d, f11, roundkeys[49], 8);
	fround (c, d, a, b, f11, roundkeys[50], 10);
	fround (d, a, b, c, f11, roundkeys[51], 14);
	fround (a, b, d, c, f11, roundkeys[52], 11);
	fround (b, c, d, a, f11, roundkeys[53], 9);
	fround (c, d, b, a, f11, roundkeys[54], 12);
	fround (d, a, c, b, f11, roundkeys[55], 7);

	fround (a, b, c, d, f10, roundkeys[40], 13);
	fround (b, c, a, d, f10, roundkeys[41], 11);
	fround (c, d, a, b, f10, roundkeys[42], 15);
	fround (d, a, b, c, f10, roundkeys[43], 8);
	fround (a, b, d, c, f10, roundkeys[44], 10);
	fround (b, c, d, a, f10, roundkeys[45], 12);
	fround (c, d, b, a, f10, roundkeys[46], 14);
	fround (d, a, c, b, f10, roundkeys[47], 9);

	fround (a, b, c, d, f9, roundkeys[36], 14);
	fround (b, c, a, d, f9, roundkeys[37], 12);
	fround (c, d, a, b, f9, roundkeys[38], 13);
	fround (d, a, b, c, f9, roundkeys[39], 11);

	fround (a, b, c, d, f8, roundkeys[32], 7);
	fround (b, c, a, d, f8, roundkeys[33], 9);
	fround (c, d, a, b, f8, roundkeys[34], 8);
	fround (d, a, b, c, f8, roundkeys[35], 10);

	fround (a, b, c, d, f7, roundkeys[28], 11);
	fround (b, c, a, d, f7, roundkeys[29], 13);
	fround (c, d, a, b, f7, roundkeys[30], 12);
	fround (d, a, b, c, f7, roundkeys[31], 14);

	fround (a, b, c, d, f6, roundkeys[24], 7);
	fround (b, c, a, d, f6, roundkeys[25], 10);
	fround (c, d, a, b, f6, roundkeys[26], 9);
	fround (d, a, b, c, f6, roundkeys[27], 8);

	fround (a, b, c, d, f5, roundkeys[20], 10);
	fround (b, c, a, d, f5, roundkeys[21], 12);
	fround (c, d, a, b, f5, roundkeys[22], 13);
	fround (d, a, b, c, f5, roundkeys[23], 11);

	fround (a, b, c, d, f4, roundkeys[16], 6);
	fround (b, c, a, d, f4, roundkeys[17], 9);
	fround (c, d, a, b, f4, roundkeys[18], 8);
	fround (d, a, b, c, f4, roundkeys[19], 7);

	fround (a, b, c, d, f3, roundkeys[12], 12);
	fround (b, c, a, d, f3, roundkeys[13], 10);
	fround (c, d, a, b, f3, roundkeys[14], 11);
	fround (d, a, b, c, f3, roundkeys[15], 9);

	fround (a, b, c, d, f2, roundkeys[8], 5);
	fround (b, c, a, d, f2, roundkeys[9], 8);
	fround (c, d, a, b, f2, roundkeys[10], 7);
	fround (d, a, b, c, f2, roundkeys[11], 6);

	fround (a, b, c, d, f1, roundkeys[4], 10);
	fround (b, c, a, d, f1, roundkeys[5], 12);
	fround (c, d, a, b, f1, roundkeys[6], 11);
	fround (d, a, b, c, f1, roundkeys[7], 9);

	fround (a, b, c, d, f0, roundkeys[0], 5);
	fround (b, c, a, d, f0, roundkeys[1], 7);
	fround (c, d, a, b, f0, roundkeys[2], 6);
	fround (d, a, b, c, f0, roundkeys[3], 8);


	data[0] = a ^ rawkey[4];
	data[1] = b ^ rawkey[5];
	data[2] = c ^ rawkey[6];
	data[3] = d ^ rawkey[7];


}


void __stdcall
decrypt (u32 * data)
{

	u32 register a, b, c, d;

	a = data[0] ^ rawkey[4];
	b = data[1] ^ rawkey[5];
	c = data[2] ^ rawkey[6];
	d = data[3] ^ rawkey[7];


	bround (d, a, b, c, f0, roundkeys[3], 8);
	bround (c, d, a, b, f0, roundkeys[2], 6);
	bround (b, c, a, d, f0, roundkeys[1], 7);
	bround (a, b, c, d, f0, roundkeys[0], 5);

	bround (d, a, b, c, f1, roundkeys[7], 9);
	bround (c, d, a, b, f1, roundkeys[6], 11);
	bround (b, c, a, d, f1, roundkeys[5], 12);
	bround (a, b, c, d, f1, roundkeys[4], 10);

	bround (d, a, b, c, f2, roundkeys[11], 6);
	bround (c, d, a, b, f2, roundkeys[10], 7);
	bround (b, c, a, d, f2, roundkeys[9], 8);
	bround (a, b, c, d, f2, roundkeys[8], 5);

	bround (d, a, b, c, f3, roundkeys[15], 9);
	bround (c, d, a, b, f3, roundkeys[14], 11);
	bround (b, c, a, d, f3, roundkeys[13], 10);
	bround (a, b, c, d, f3, roundkeys[12], 12);

	bround (d, a, b, c, f4, roundkeys[19], 7);
	bround (c, d, a, b, f4, roundkeys[18], 8);
	bround (b, c, a, d, f4, roundkeys[17], 9);
	bround (a, b, c, d, f4, roundkeys[16], 6);

	bround (d, a, b, c, f5, roundkeys[23], 11);
	bround (c, d, a, b, f5, roundkeys[22], 13);
	bround (b, c, a, d, f5, roundkeys[21], 12);
	bround (a, b, c, d, f5, roundkeys[20], 10);

	bround (d, a, b, c, f6, roundkeys[27], 8);
	bround (c, d, a, b, f6, roundkeys[26], 9);
	bround (b, c, a, d, f6, roundkeys[25], 10);
	bround (a, b, c, d, f6, roundkeys[24], 7);

	bround (d, a, b, c, f7, roundkeys[31], 14);
	bround (c, d, a, b, f7, roundkeys[30], 12);
	bround (b, c, a, d, f7, roundkeys[29], 13);
	bround (a, b, c, d, f7, roundkeys[28], 11);

	bround (d, a, b, c, f8, roundkeys[35], 10);
	bround (c, d, a, b, f8, roundkeys[34], 8);
	bround (b, c, a, d, f8, roundkeys[33], 9);
	bround (a, b, c, d, f8, roundkeys[32], 7);

	bround (d, a, b, c, f9, roundkeys[39], 11);
	bround (c, d, a, b, f9, roundkeys[38], 13);
	bround (b, c, a, d, f9, roundkeys[37], 12);
	bround (a, b, c, d, f9, roundkeys[36], 14);

	bround (d, a, c, b, f10, roundkeys[47], 9);
	bround (c, d, b, a, f10, roundkeys[46], 14);
	bround (b, c, d, a, f10, roundkeys[45], 12);
	bround (a, b, d, c, f10, roundkeys[44], 10);
	bround (d, a, b, c, f10, roundkeys[43], 8);
	bround (c, d, a, b, f10, roundkeys[42], 15);
	bround (b, c, a, d, f10, roundkeys[41], 11);
	bround (a, b, c, d, f10, roundkeys[40], 13);

	bround (d, a, c, b, f11, roundkeys[55], 7);
	bround (c, d, b, a, f11, roundkeys[54], 12);
	bround (b, c, d, a, f11, roundkeys[53], 9);
	bround (a, b, d, c, f11, roundkeys[52], 11);
	bround (d, a, b, c, f11, roundkeys[51], 14);
	bround (c, d, a, b, f11, roundkeys[50], 10);
	bround (b, c, a, d, f11, roundkeys[49], 8);
	bround (a, b, c, d, f11, roundkeys[48], 13);

	bround (d, a, c, b, f12, roundkeys[63], 11);
	bround (c, d, b, a, f12, roundkeys[62], 6);
	bround (b, c, d, a, f12, roundkeys[61], 13);
	bround (a, b, d, c, f12, roundkeys[60], 9);
	bround (d, a, b, c, f12, roundkeys[59], 10);
	bround (c, d, a, b, f12, roundkeys[58], 8);
	bround (b, c, a, d, f12, roundkeys[57], 7);
	bround (a, b, c, d, f12, roundkeys[56], 12);

	bround (d, a, c, b, f13, roundkeys[71], 6);
	bround (c, d, b, a, f13, roundkeys[70], 10);
	bround (b, c, d, a, f13, roundkeys[69], 5);
	bround (a, b, d, c, f13, roundkeys[68], 9);
	bround (d, a, b, c, f13, roundkeys[67], 8);
	bround (c, d, a, b, f13, roundkeys[66], 11);
	bround (b, c, a, d, f13, roundkeys[65], 7);
	bround (a, b, c, d, f13, roundkeys[64], 12);


	bround (d, a, c, b, f14, roundkeys[79], 4);
	bround (c, d, b, a, f14, roundkeys[78], 11);
	bround (b, c, d, a, f14, roundkeys[77], 8);
	bround (a, b, d, c, f14, roundkeys[76], 5);
	bround (d, a, b, c, f14, roundkeys[75], 7);
	bround (c, d, a, b, f14, roundkeys[74], 5);
	bround (b, c, a, d, f14, roundkeys[73], 9);
	bround (a, b, c, d, f14, roundkeys[72], 10);

	bround (d, a, c, b, f15, roundkeys[87], 14);
	bround (c, d, b, a, f15, roundkeys[86], 9);
	bround (b, c, d, a, f15, roundkeys[85], 11);
	bround (a, b, d, c, f15, roundkeys[84], 10);
	bround (d, a, b, c, f15, roundkeys[83], 13);
	bround (c, d, a, b, f15, roundkeys[82], 12);
	bround (b, c, a, d, f15, roundkeys[81], 15);
	bround (a, b, c, d, f15, roundkeys[80], 8);

	bround (d, a, b, c, f3, roundkeys[91], 6);
	bround (c, d, a, b, f2, roundkeys[90], 8);
	bround (b, c, a, d, f1, roundkeys[89], 5);
	bround (a, b, c, d, f0, roundkeys[88], 7);

	bround (d, a, b, c, f7, roundkeys[95], 5);
	bround (c, d, a, b, f6, roundkeys[94], 7);
	bround (b, c, a, d, f5, roundkeys[93], 8);
	bround (a, b, c, d, f4, roundkeys[92], 6);

	bround (d, a, b, c, f0, roundkeys[99], 6);
	bround (c, d, a, b, f1, roundkeys[98], 8);
	bround (b, c, a, d, f2, roundkeys[97], 7);
	bround (a, b, c, d, f3, roundkeys[96], 5);

	bround (d, a, b, c, f4, roundkeys[103], 6);
	bround (c, d, a, b, f5, roundkeys[102], 7);
	bround (b, c, a, d, f6, roundkeys[101], 8);
	bround (a, b, c, d, f7, roundkeys[100], 5);

	bround (d, a, c, b, f15, roundkeys[111], 14);
	bround (c, d, b, a, f14, roundkeys[110], 10);
	bround (b, c, d, a, f13, roundkeys[109], 11);
	bround (a, b, d, c, f12, roundkeys[108], 12);
	bround (d, a, b, c, f11, roundkeys[107], 15);
	bround (c, d, a, b, f10, roundkeys[106], 13);
	bround (b, c, a, d, f9, roundkeys[105], 9);
	bround (a, b, c, d, f8, roundkeys[104], 8);

	bround (d, a, c, b, f7, roundkeys[119], 9);
	bround (c, d, b, a, f6, roundkeys[118], 15);
	bround (b, c, d, a, f5, roundkeys[117], 10);
	bround (a, b, d, c, f4, roundkeys[116], 12);
	bround (d, a, b, c, f3, roundkeys[115], 13);
	bround (c, d, a, b, f2, roundkeys[114], 11);
	bround (b, c, a, d, f1, roundkeys[113], 9);
	bround (a, b, c, d, f0, roundkeys[112], 14);

	bround (d, a, c, b, f8, roundkeys[127], 8);
	bround (c, d, b, a, f9, roundkeys[126], 12);
	bround (b, c, d, a, f10, roundkeys[125], 13);
	bround (a, b, d, c, f11, roundkeys[124], 15);
	bround (d, a, b, c, f12, roundkeys[123], 10);
	bround (c, d, a, b, f13, roundkeys[122], 11);
	bround (b, c, a, d, f14, roundkeys[121], 9);
	bround (a, b, c, d, f15, roundkeys[120], 14);


	data[0] = a ^ rawkey[0];
	data[1] = b ^ rawkey[1];
	data[2] = c ^ rawkey[2];
	data[3] = d ^ rawkey[3];



}


#ifdef PRNG_TEST

u32 buf[4] = { 0xC4A60B29, 0x1DF6907A, 0xE5F8D137, 0x86C02E59 };
u32 cbuf[4] = { 0xBFE60D0F, 0x7CCB60D6, 0x926A9747, 0x17EE621F };
u32 key[16] =
{ 0x58819022, 0xAC3F8B9D, 0x9CEDAECB, 0x89EA9D3D, 0xB43CD725, 0xF04F4E02,
0x12DF1FF7, 0xDBED2295, 0x9140703B, 0x63C99153, 0x786F5204, 0xF0DEA408,
0x3BE2517F, 0xEB12063E, 0xBEC1F2A8, 0x3A498DC2
};

int
main (int argc, char *argv[])
{
	FILE *f;
	int i, n;

	if (argc < 2)
	{
		fprintf (stderr, "usage: %s <output filename>\n", argv[0]);
		return (1);
	}
	if ((f = fopen (argv[1], "wb+")) == NULL)
	{
		fprintf (stderr, "failed to open '%s' for writing.\n", argv[1]);
		return (1);
	}



	setup (&key[0]);
	for (i = 0; i != 67108864; i++)
	{

		crypt (&buf[0]);

		for (n = 0; n != 4; n++)
		{
			cbuf[n] ^= buf[n];
		}

		fwrite (cbuf, 16, 1, f);


	}

	fclose (f);
	return (0);
}

#endif



u32 __stdcall getblocksize ()
{
	return 128;
}

u32 __stdcall getkeysize ()
{
	return keysizebits;
}

void __stdcall getciphername (unsigned char *p)
{
strcpy(p, "Sonjitege");
}