/********************************************************************

Umchak - experimental symmetric block cipher

Runs with 256..4096 bit key and 64 bit block.

Edition 1.1 dated: 11.03.2009

(c) 2009 Alexander Myasnikov

Web: www.alexanderwdark.narod.ru

********************************************************************/

#include <stdlib.h>
#include <string.h>

#ifdef PRNG_TEST
#include <stdio.h>
#endif

typedef unsigned char u8;
typedef unsigned long u32;
typedef unsigned __int64 u64;
typedef unsigned short u16;

typedef long int s32;

/************** Full key size 256 - 4096 ***************************/
#define keysizebits  512
/*******************************************************************/

static const u32 keysizebytes = keysizebits / 8;
static const u32 keysize = keysizebytes / 4;

/************** Number of algorithm full cycles 16/32/48/64 ********/
#define cycles 16
/*******************************************************************/

/************** Sbox type 1 - 7 ************************************/
#define sizeshift 4
/*******************************************************************/

/************** How to shift in F function  ************************/
#define SHLVAL 6
#define SHRVAL 8
/*******************************************************************/

/************** Optimal constants for key expand  ******************/
#define delta  0xb7e15163
#define cdelta 0x9e3779b9
/*******************************************************************/

/************** How to expand key **********************************/
#define USE_EXTRA_KS
/*******************************************************************/


static const u32 sboxsize = (2 << sizeshift);

static const u32 rounds = cycles << 2;

#include "sbox.h"

#define ROL(x,y) _lrotl((x), (y))
#define ROR(x,y) _lrotr((x), (y))

#define subst(x,a,b,c,d, r) (ROL((sbox[a][(x) >> 24 & 255] << 24 | sbox[b][(x) >> 16 & 255] << 16 |  sbox[c][(x) >> 8 & 255] << 8 | sbox[d][(x) & 255]),(r)))

#define BSWAP(x) (((x) << 16) | ((x) >> 16))
#define F(x) (BSWAP(x) + ((x << SHLVAL) ^ (x >> SHRVAL)))
#define FQ(x, y) (((x) ^ (y)) + ((x << SHLVAL) ^ (x >> SHRVAL)))

u32 _roundkeys[rounds];
u32 roundkeys[rounds];

u32 rawkey[8];

void __stdcall
setup (u32 * key)
{

#ifdef USE_EXTRA_KS
	u8 *subkeys = (u8 *) & roundkeys[0];
	u8 *material = (u8 *) & _roundkeys[0];
	u8 tmp;
#endif

	u32 cnt = delta, kcnt = cdelta, i, n;
	u32 rk[keysize];


	for (i = 0; i != keysize; i++)
	{
		rk[i] = key[i];
	}

	for (i = 0; i != rounds; i++)
	{

		_roundkeys[i] = rk[i % keysize];
		kcnt += rk[i % keysize] + cdelta;
		_roundkeys[i] += cnt;
		_roundkeys[i] =
			subst (_roundkeys[i], i % sboxsize, cnt % sboxsize,
			_roundkeys[i] % sboxsize, kcnt % sboxsize, 11);
		kcnt += _roundkeys[i];
		_roundkeys[i] = F (_roundkeys[i]) + rk[i % keysize];
		_roundkeys[i] =
			subst (_roundkeys[i], kcnt % sboxsize, _roundkeys[i] % sboxsize,
			cnt % sboxsize, i % sboxsize, 11);
		kcnt += _roundkeys[i];
		cnt += delta;

	}

#ifdef USE_EXTRA_KS

	cnt = 0;

	for (n = 0; n != rounds >> 6; n++)
	{

		for (i = 0; i != 256; i++)
		{
			subkeys[i] = i;
		}

		for (i = 0; i != 768; i++)
		{
			cnt = subkeys[(cnt + subkeys[i & 0xff] + material[i % 256]) & 0xff];
			tmp = subkeys[i & 0xff];
			subkeys[i & 0xff] = subkeys[cnt & 0xff];
			subkeys[cnt & 0xff] = tmp;
		}

		subkeys += 0x100;
		material += 0x100;
	}

#else

	for (i = 0; i != rounds; i++)
	{

		roundkeys[i] = _roundkeys[i];
	}

#endif


	for (i = 0; i != 8; i++)
	{
		rawkey[i] = rk[i];
	}

	if (keysize >= 16)

	{
		for (i = 8; i != 16; i++)

		{

			rawkey[i % 8] += rk[i];

		}

	}

	if (keysize > 16)

	{
		for (i = 16; i != keysize; i++)

		{

			rawkey[i % 8] ^= rk[i];

		}
	}

}





void __stdcall
crypt (u32 * data)
{
	register u32 a, b;

	u32 i = 0;

	a = data[0] + roundkeys[i++];
	b = data[1] + roundkeys[i++];

	for (; i != rounds - 2;)
	{

		a = ROL (a + FQ (b, i), b & 31) + roundkeys[i++];
		b = ROL (b + FQ (a, i), a & 31) + roundkeys[i++];

		a = ROL (a + FQ (b, i), b >> 27) + roundkeys[i++];
		b = ROL (b + FQ (a, i), a >> 27) + roundkeys[i++];

	}

	data[0] = a ^ roundkeys[i++];
	data[1] = b ^ roundkeys[i++];

}


void __stdcall
decrypt (u32 * data)
{
	register u32 a, b;
	u32 i = rounds - 1;

	b = data[1] ^ roundkeys[i--];
	a = data[0] ^ roundkeys[i--];


	for (; i != 1;)

	{

		b = ROR (b - roundkeys[i], a >> 27) - FQ (a, i--);
		a = ROR (a - roundkeys[i], b >> 27) - FQ (b, i--);

		b = ROR (b - roundkeys[i], a & 31) - FQ (a, i--);
		a = ROR (a - roundkeys[i], b & 31) - FQ (b, i--);


	}


	data[1] = b - roundkeys[i--];
	data[0] = a - roundkeys[i--];

}


#ifdef PRNG_TEST

u32 buf[2] = { 0xC4A60B29, 0x1DF6907A };
u32 cbuf[2] = { 0xBFE60D0F, 0x7CCB60D6 };
u32 key[16] =
{ 0x58819022, 0xAC3F8B9D, 0x9CEDAECB, 0x89EA9D3D, 0xB43CD725, 0xF04F4E02,
0x12DF1FF7, 0xDBED2295, 0x9140703B, 0x63C99153, 0x786F5204, 0xF0DEA408,
0x3BE2517F, 0xEB12063E, 0xBEC1F2A8, 0x3A498DC2
};

int
main (int argc, char *argv[])
{
	FILE *f;
	int i, n;

	if (argc < 2)
	{
		fprintf (stderr, "usage: %s <output filename>\n", argv[0]);
		return (1);
	}
	if ((f = fopen (argv[1], "wb+")) == NULL)
	{
		fprintf (stderr, "failed to open '%s' for writing.\n", argv[1]);
		return (1);
	}



	setup (&key[0]);
	for (i = 0; i != 134217728; i++)
	{

		crypt (&buf[0]);

		for (n = 0; n != 2; n++)
		{
			cbuf[n] ^= buf[n];
		}

		fwrite (cbuf, 8, 1, f);


	}

	fclose (f);
	return (0);
}

#endif


u32 __stdcall getblocksize ()
{
	return 64;
}

u32 __stdcall getkeysize ()
{
	return keysizebits;
}

void __stdcall getciphername (unsigned char *p)
{
strcpy(p, "Umchak");
}