/********************************************************************

Webino - experimental software symmetric block cipher

Runs with 256..2048 bit key and 128 bit block.

Edition 1.0 dated: 21.03.2009

(C) 2009 Alexander Myasnikov, public domain

Web: www.alexanderwdark.narod.ru

********************************************************************/

#include <stdlib.h>
#include <string.h>

#ifdef PRNG_TEST
#include <stdio.h>
#endif

typedef unsigned char u8;
typedef unsigned long u32;
typedef unsigned short u16;

/************** Full key size 256 - 4096 ***************************/
#define keysizebits  512
/*******************************************************************/

static const u32 keysizebytes = keysizebits / 8;
static const u32 keysize = keysizebytes / 4;

/************** Sbox type 1 - 7 ************************************/
#define sizeshift 5
/*******************************************************************/

/************** How to shift in F function  ************************/
#define SHLVAL 6
#define SHRVAL 8
/*******************************************************************/

/************** Optimal constants for key expand  ******************/
#define delta  0xb7e15163
#define cdelta 0x9e3779b9
/*******************************************************************/

/************** How to expand key **********************************/
#define USE_EXTRA_KS
/*******************************************************************/


#define rounds 64

static const u32 sboxsize = (2 << sizeshift);

#include "sbox.h"

#define ROL(x,y) _lrotl((x), (y))
#define ROR(x,y) _lrotr((x), (y))

#define subst(x,a,b,c,d, r) (ROL((sbox[a][(x) >> 24 & 255] << 24 | sbox[b][(x) >> 16 & 255] << 16 |  sbox[c][(x) >> 8 & 255] << 8 | sbox[d][(x) & 255]),(r)))

#define BSWAP(x) (((x) << 16) | ((x) >> 16))

#define FQ(x, y) ((x ^ y) + ((x << SHLVAL) ^ (x >> SHRVAL)))
#define F(x) (BSWAP(x) + ((x << SHLVAL) ^ (x >> SHRVAL)))
#define byte(x,n)   ((u8)((x) >> (8 * n)))

u32 roundkeys[rounds], _roundkeys[rounds];

u32 rawkey[8];


void __stdcall
setup (u32 * key)
{

#ifdef USE_EXTRA_KS
	u8 *subkeys = (u8 *) & roundkeys[0];
	u8 *material = (u8 *) & _roundkeys[0];
	u8 tmp;
#endif

	u32 cnt = delta, kcnt = cdelta, i, n;
	u32 rk[keysize];


	for (i = 0; i != keysize; i++)
	{
		rk[i] = key[i];
	}

	for (i = 0; i != rounds; i++)
	{

		_roundkeys[i] = rk[i % keysize];
		kcnt += rk[i % keysize] + cdelta;
		_roundkeys[i] += cnt;
		_roundkeys[i] =
			subst (_roundkeys[i], i % sboxsize, cnt % sboxsize,
			_roundkeys[i] % sboxsize, kcnt % sboxsize, 11);
		kcnt += _roundkeys[i];
		_roundkeys[i] = F (_roundkeys[i]) + rk[i % keysize];
		_roundkeys[i] =
			subst (_roundkeys[i], kcnt % sboxsize, _roundkeys[i] % sboxsize,
			cnt % sboxsize, i % sboxsize, 11);
		kcnt += _roundkeys[i];
		cnt += delta;

	}

#ifdef USE_EXTRA_KS

	cnt = 0;

	for (n = 0; n != rounds >> 6; n++)
	{

		for (i = 0; i != 256; i++)
		{
			subkeys[i] = i;
		}

		for (i = 0; i != 768; i++)
		{
			cnt = subkeys[(cnt + subkeys[i & 0xff] + material[i % 256]) & 0xff];
			tmp = subkeys[i & 0xff];
			subkeys[i & 0xff] = subkeys[cnt & 0xff];
			subkeys[cnt & 0xff] = tmp;
		}

		subkeys += 0x100;
		material += 0x100;
	}

#else

	for (i = 0; i != rounds; i++)
	{

		roundkeys[i] = _roundkeys[i];
	}

#endif



	for (i = 0; i != 8; i++)
	{
		rawkey[i] = rk[i];
	}

	if (keysize >= 16)

	{
		for (i = 8; i != 16; i++)

		{

			rawkey[i % 8] += rk[i];

		}

	}

	if (keysize > 16)

	{
		for (i = 16; i != keysize; i++)

		{

			rawkey[i % 8] ^= rk[i];

		}
	}


}

#define mdF(x, y, z) (((x) & (y)) | ((~x) & (z)))
#define G(x, y, z) (((x) & (z)) | ((y) & (~z)))
#define H(x, y, z) ((x) ^ (y) ^ (z))
#define I(x, y, z) ((y) ^ ((x) | (~z)))

#define FF(a, b, c, d, x, s, ac)  (a) += mdF ((b), (c), (d)) + (x) + (u32)(ac); (a) = ROL ((a), (s)); (a) += (b);
#define GG(a, b, c, d, x, s, ac) (a) += G ((b), (c), (d)) + (x) + (u32)(ac); (a) = ROL ((a), (s)); (a) += (b);
#define HH(a, b, c, d, x, s, ac) (a) += H ((b), (c), (d)) + (x) + (u32)(ac); (a) = ROL ((a), (s)); (a) += (b);
#define II(a, b, c, d, x, s, ac) (a) += I ((b), (c), (d)) + (x) + (u32)(ac); (a) = ROL ((a), (s)); (a) += (b);
#define BFF(a, b, c, d, x, s, ac) (a) -= (b); (a) = ROR ((a), (s));  (a) -= mdF ((b), (c), (d)) + (x) + (u32)(ac);
#define BGG(a, b, c, d, x, s, ac) (a) -= (b);  (a) = ROR ((a), (s)); (a) -= G ((b), (c), (d)) + (x) + (u32)(ac);
#define BHH(a, b, c, d, x, s, ac)  (a) -= (b); (a) = ROR ((a), (s));  (a) -= H ((b), (c), (d)) + (x) + (u32)(ac);
#define BII(a, b, c, d, x, s, ac) (a) -= (b); (a) = ROR ((a), (s)); (a) -= I ((b), (c), (d)) + (x) + (u32)(ac);


#define S11 7
#define S12 12
#define S13 17
#define S14 22
#define S21 5
#define S22 9
#define S23 14
#define S24 20
#define S31 4
#define S32 11
#define S33 16
#define S34 23
#define S41 6
#define S42 10
#define S43 15
#define S44 21

void __stdcall
crypt (u32 * data)
{
	u32 register a, b, c, d;

	a = data[0] + rawkey[0];
	b = data[1] + rawkey[1];
	c = data[2] + rawkey[2];
	d = data[3] + rawkey[3];

	FF (a, b, c, d, roundkeys[0], S11, 3614090360);	
	FF (d, a, b, c, roundkeys[1], S12, 3905402710);	
	FF (c, d, a, b, roundkeys[2], S13, 606105819);	
	FF (b, c, d, a, roundkeys[3], S14, 3250441966);	
	FF (a, b, c, d, roundkeys[4], S11, 4118548399);	
	FF (d, a, b, c, roundkeys[5], S12, 1200080426);	
	FF (c, d, a, b, roundkeys[6], S13, 2821735955);	
	FF (b, c, d, a, roundkeys[7], S14, 4249261313);	
	FF (a, b, c, d, roundkeys[8], S11, 1770035416);	
	FF (d, a, b, c, roundkeys[9], S12, 2336552879);	
	FF (c, d, a, b, roundkeys[10], S13, 4294925233);
	FF (b, c, d, a, roundkeys[11], S14, 2304563134);
	FF (a, b, c, d, roundkeys[12], S11, 1804603682);
	FF (d, a, b, c, roundkeys[13], S12, 4254626195);
	FF (c, d, a, b, roundkeys[14], S13, 2792965006);
	FF (b, c, d, a, roundkeys[15], S14, 1236535329);

	GG (a, b, c, d, roundkeys[16], S21, 4129170786);
	GG (d, a, b, c, roundkeys[17], S22, 3225465664);
	GG (c, d, a, b, roundkeys[18], S23, 643717713);	
	GG (b, c, d, a, roundkeys[19], S24, 3921069994);
	GG (a, b, c, d, roundkeys[20], S21, 3593408605);
	GG (d, a, b, c, roundkeys[21], S22, 38016083);	
	GG (c, d, a, b, roundkeys[22], S23, 3634488961);
	GG (b, c, d, a, roundkeys[23], S24, 3889429448);
	GG (a, b, c, d, roundkeys[24], S21, 568446438);	
	GG (d, a, b, c, roundkeys[25], S22, 3275163606);
	GG (c, d, a, b, roundkeys[26], S23, 4107603335);
	GG (b, c, d, a, roundkeys[27], S24, 1163531501);
	GG (a, b, c, d, roundkeys[28], S21, 2850285829);
	GG (d, a, b, c, roundkeys[29], S22, 4243563512);
	GG (c, d, a, b, roundkeys[30], S23, 1735328473);
	GG (b, c, d, a, roundkeys[31], S24, 2368359562);

	HH (a, b, c, d, roundkeys[32], S31, 4294588738);
	HH (d, a, b, c, roundkeys[33], S32, 2272392833);
	HH (c, d, a, b, roundkeys[34], S33, 1839030562);
	HH (b, c, d, a, roundkeys[35], S34, 4259657740);
	HH (a, b, c, d, roundkeys[36], S31, 2763975236);
	HH (d, a, b, c, roundkeys[37], S32, 1272893353);
	HH (c, d, a, b, roundkeys[38], S33, 4139469664);
	HH (b, c, d, a, roundkeys[39], S34, 3200236656);
	HH (a, b, c, d, roundkeys[40], S31, 681279174);	
	HH (d, a, b, c, roundkeys[41], S32, 3936430074);
	HH (c, d, a, b, roundkeys[42], S33, 3572445317);
	HH (b, c, d, a, roundkeys[43], S34, 76029189);	
	HH (a, b, c, d, roundkeys[44], S31, 3654602809);
	HH (d, a, b, c, roundkeys[45], S32, 3873151461);
	HH (c, d, a, b, roundkeys[46], S33, 530742520);	
	HH (b, c, d, a, roundkeys[47], S34, 3299628645);

	II (a, b, c, d, roundkeys[48], S41, 4096336452);
	II (d, a, b, c, roundkeys[49], S42, 1126891415);
	II (c, d, a, b, roundkeys[50], S43, 2878612391);
	II (b, c, d, a, roundkeys[51], S44, 4237533241);
	II (a, b, c, d, roundkeys[52], S41, 1700485571);
	II (d, a, b, c, roundkeys[53], S42, 2399980690);
	II (c, d, a, b, roundkeys[54], S43, 4293915773);
	II (b, c, d, a, roundkeys[55], S44, 2240044497);
	II (a, b, c, d, roundkeys[56], S41, 1873313359);
	II (d, a, b, c, roundkeys[57], S42, 4264355552);
	II (c, d, a, b, roundkeys[58], S43, 2734768916);
	II (b, c, d, a, roundkeys[59], S44, 1309151649);
	II (a, b, c, d, roundkeys[60], S41, 4149444226);
	II (d, a, b, c, roundkeys[61], S42, 3174756917);
	II (c, d, a, b, roundkeys[62], S43, 718787259);	
	II (b, c, d, a, roundkeys[63], S44, 3951481745);


	data[0] = a ^ rawkey[4];
	data[1] = b ^ rawkey[5];
	data[2] = c ^ rawkey[6];
	data[3] = d ^ rawkey[7];


}


void __stdcall
decrypt (u32 * data)
{
	u32 register a, b, c, d;

	a = data[0] ^ rawkey[4];
	b = data[1] ^ rawkey[5];
	c = data[2] ^ rawkey[6];
	d = data[3] ^ rawkey[7];

	BII (b, c, d, a, roundkeys[63], S44, 3951481745);
	BII (c, d, a, b, roundkeys[62], S43, 718787259);
	BII (d, a, b, c, roundkeys[61], S42, 3174756917);
	BII (a, b, c, d, roundkeys[60], S41, 4149444226);
	BII (b, c, d, a, roundkeys[59], S44, 1309151649);
	BII (c, d, a, b, roundkeys[58], S43, 2734768916);
	BII (d, a, b, c, roundkeys[57], S42, 4264355552);
	BII (a, b, c, d, roundkeys[56], S41, 1873313359);
	BII (b, c, d, a, roundkeys[55], S44, 2240044497);
	BII (c, d, a, b, roundkeys[54], S43, 4293915773);
	BII (d, a, b, c, roundkeys[53], S42, 2399980690);
	BII (a, b, c, d, roundkeys[52], S41, 1700485571);
	BII (b, c, d, a, roundkeys[51], S44, 4237533241);
	BII (c, d, a, b, roundkeys[50], S43, 2878612391);
	BII (d, a, b, c, roundkeys[49], S42, 1126891415);
	BII (a, b, c, d, roundkeys[48], S41, 4096336452);

	BHH (b, c, d, a, roundkeys[47], S34, 3299628645);
	BHH (c, d, a, b, roundkeys[46], S33, 530742520);
	BHH (d, a, b, c, roundkeys[45], S32, 3873151461);
	BHH (a, b, c, d, roundkeys[44], S31, 3654602809);
	BHH (b, c, d, a, roundkeys[43], S34, 76029189);	
	BHH (c, d, a, b, roundkeys[42], S33, 3572445317);
	BHH (d, a, b, c, roundkeys[41], S32, 3936430074);
	BHH (a, b, c, d, roundkeys[40], S31, 681279174);
	BHH (b, c, d, a, roundkeys[39], S34, 3200236656);
	BHH (c, d, a, b, roundkeys[38], S33, 4139469664);
	BHH (d, a, b, c, roundkeys[37], S32, 1272893353);
	BHH (a, b, c, d, roundkeys[36], S31, 2763975236);
	BHH (b, c, d, a, roundkeys[35], S34, 4259657740);
	BHH (c, d, a, b, roundkeys[34], S33, 1839030562);
	BHH (d, a, b, c, roundkeys[33], S32, 2272392833);
	BHH (a, b, c, d, roundkeys[32], S31, 4294588738);

	BGG (b, c, d, a, roundkeys[31], S24, 2368359562);
	BGG (c, d, a, b, roundkeys[30], S23, 1735328473);
	BGG (d, a, b, c, roundkeys[29], S22, 4243563512);
	BGG (a, b, c, d, roundkeys[28], S21, 2850285829);
	BGG (b, c, d, a, roundkeys[27], S24, 1163531501);
	BGG (c, d, a, b, roundkeys[26], S23, 4107603335);
	BGG (d, a, b, c, roundkeys[25], S22, 3275163606);
	BGG (a, b, c, d, roundkeys[24], S21, 568446438);
	BGG (b, c, d, a, roundkeys[23], S24, 3889429448);
	BGG (c, d, a, b, roundkeys[22], S23, 3634488961);
	BGG (d, a, b, c, roundkeys[21], S22, 38016083);	
	BGG (a, b, c, d, roundkeys[20], S21, 3593408605);
	BGG (b, c, d, a, roundkeys[19], S24, 3921069994);
	BGG (c, d, a, b, roundkeys[18], S23, 643717713);
	BGG (d, a, b, c, roundkeys[17], S22, 3225465664);
	BGG (a, b, c, d, roundkeys[16], S21, 4129170786);

	BFF (b, c, d, a, roundkeys[15], S14, 1236535329);
	BFF (c, d, a, b, roundkeys[14], S13, 2792965006);
	BFF (d, a, b, c, roundkeys[13], S12, 4254626195);
	BFF (a, b, c, d, roundkeys[12], S11, 1804603682);
	BFF (b, c, d, a, roundkeys[11], S14, 2304563134);
	BFF (c, d, a, b, roundkeys[10], S13, 4294925233);
	BFF (d, a, b, c, roundkeys[9], S12, 2336552879);
	BFF (a, b, c, d, roundkeys[8], S11, 1770035416);
	BFF (b, c, d, a, roundkeys[7], S14, 4249261313);
	BFF (c, d, a, b, roundkeys[6], S13, 2821735955);
	BFF (d, a, b, c, roundkeys[5], S12, 1200080426);
	BFF (a, b, c, d, roundkeys[4], S11, 4118548399);
	BFF (b, c, d, a, roundkeys[3], S14, 3250441966);
	BFF (c, d, a, b, roundkeys[2], S13, 606105819);	
	BFF (d, a, b, c, roundkeys[1], S12, 3905402710);
	BFF (a, b, c, d, roundkeys[0], S11, 3614090360);

	data[0] = a - rawkey[0];
	data[1] = b - rawkey[1];
	data[2] = c - rawkey[2];
	data[3] = d - rawkey[3];

}


#ifdef PRNG_TEST

u32 buf[4] = { 0xC4A60B29, 0x1DF6907A, 0xE5F8D137, 0x86C02E59 };
u32 cbuf[4] = { 0xBFE60D0F, 0x7CCB60D6, 0x926A9747, 0x17EE621F };
u32 key[16] =
{ 0x58819022, 0xAC3F8B9D, 0x9CEDAECB, 0x89EA9D3D, 0xB43CD725, 0xF04F4E02,
0x12DF1FF7, 0xDBED2295, 0x9140703B, 0x63C99153, 0x786F5204, 0xF0DEA408,
0x3BE2517F, 0xEB12063E, 0xBEC1F2A8, 0x3A498DC2
};

int
main (int argc, char *argv[])
{
	FILE *f;
	int i, n;

	if (argc < 2)
	{
		fprintf (stderr, "usage: %s <output filename>\n", argv[0]);
		return (1);
	}
	if ((f = fopen (argv[1], "wb+")) == NULL)
	{
		fprintf (stderr, "failed to open '%s' for writing.\n", argv[1]);
		return (1);
	}



	setup (&key[0]);
	for (i = 0; i != 67108864; i++)
	{

		crypt (&buf[0]);

		for (n = 0; n != 4; n++)
		{
			cbuf[n] ^= buf[n];
		}

		fwrite (cbuf, 16, 1, f);


	}

	fclose (f);
	return (0);
}

#endif


u32 __stdcall getblocksize ()
{
	return 128;
}

u32 __stdcall getkeysize ()
{
	return keysizebits;
}

void __stdcall getciphername (unsigned char *p)
{
strcpy(p, "Webino");
}


